@@ -33,6 +33,7 @@
 #endif
 
 #ifdef G_OS_WIN32
+#include <windows.h>
 #include <errno.h>
 #include <wchar.h>
 #include <direct.h>
@@ -336,10 +337,9 @@ g_creat (const gchar *filename,
  * renames a file, moving it between directories if required.
  * 
  * See your C library manual for more details about how rename() works
- * on your system. Note in particular that on Windows, it is in
- * general not possible to rename a file if a file with the new name
- * already exists. Also it is not possible in general to rename an
- * open file.
+ * on your system. Note in particular that on Win9x it is not possible
+ * to rename a file if a file with the new name already exists. Also
+ * it is not possible in general on Windows to rename an open file.
  *
  * Returns: 0 if the renaming succeeded, -1 if an error occurred
  * 
@@ -355,7 +355,7 @@ g_rename (const gchar *oldfilename,
       wchar_t *woldfilename = g_utf8_to_utf16 (oldfilename, -1, NULL, NULL, NULL);
       wchar_t *wnewfilename;
       int retval;
-      int save_errno;
+      int save_errno = 0;
 
       if (woldfilename == NULL)
 	{
@@ -372,8 +372,26 @@ g_rename (const gchar *oldfilename,
 	  return -1;
 	}
 
-      retval = _wrename (woldfilename, wnewfilename);
-      save_errno = errno;
+      if (MoveFileExW (woldfilename, wnewfilename, MOVEFILE_REPLACE_EXISTING))
+	retval = 0;
+      else
+	{
+	  retval = -1;
+	  switch (GetLastError ())
+	    {
+#define CASE(a,b) case ERROR_##a: save_errno = b; break
+	    CASE (FILE_NOT_FOUND, ENOENT);
+	    CASE (PATH_NOT_FOUND, ENOENT);
+	    CASE (ACCESS_DENIED, EACCES);
+	    CASE (NOT_SAME_DEVICE, EXDEV);
+	    CASE (LOCK_VIOLATION, EACCES);
+	    CASE (SHARING_VIOLATION, EACCES);
+	    CASE (FILE_EXISTS, EEXIST);
+	    CASE (ALREADY_EXISTS, EEXIST);
+#undef CASE
+	    default: save_errno = EIO;
+	    }
+	}
 
       g_free (woldfilename);
       g_free (wnewfilename);
@@ -570,6 +588,7 @@ g_stat (const gchar *filename,
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
       int retval;
       int save_errno;
+      int len;
 
       if (wfilename == NULL)
 	{
@@ -577,6 +596,13 @@ g_stat (const gchar *filename,
 	  return -1;
 	}
 
+      len = wcslen (wfilename);
+      while (len > 0 && G_IS_DIR_SEPARATOR (wfilename[len-1]))
+	len--;
+      if (len > 0 &&
+	  (!g_path_is_absolute (filename) || len > g_path_skip_root (filename) - filename))
+	wfilename[len] = '\0';
+
       retval = _wstat (wfilename, (struct _stat *) buf);
       save_errno = errno;
 
@@ -590,6 +616,7 @@ g_stat (const gchar *filename,
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
       int retval;
       int save_errno;
+      int len;
 
       if (cp_filename == NULL)
 	{
@@ -597,6 +624,13 @@ g_stat (const gchar *filename,
 	  return -1;
 	}
 
+      len = strlen (cp_filename);
+      while (len > 0 && G_IS_DIR_SEPARATOR (cp_filename[len-1]))
+	len--;
+      if (len > 0 &&
+	  (!g_path_is_absolute (filename) || len > g_path_skip_root (filename) - filename))
+	cp_filename[len] = '\0';
+      
       retval = stat (cp_filename, buf);
       save_errno = errno;
 
@@ -724,6 +758,11 @@ g_unlink (const gchar *filename)
  * Windows, it is in general not possible to remove a file that is
  * open to some process, or mapped into memory.
  *
+ * If this function fails on Windows you can't infer too much from the
+ * errno value. rmdir() is tried regardless of what caused remove() to
+ * fail. Any errno value set by remove() will be overwritten by that
+ * set by rmdir().
+ *
  * Returns: 0 if the file was successfully removed, -1 if an error 
  *    occurred
  * 
@@ -857,7 +896,7 @@ g_rmdir (const gchar *filename)
  * 
  * See the C library manual for more details about fopen().
  *
- * Returns: A <typename>FILE</typename> pointer if the file was successfully
+ * Returns: A <type>FILE</type> pointer if the file was successfully
  *    opened, or %NULL if an error occurred
  * 
  * Since: 2.6
@@ -935,7 +974,7 @@ g_fopen (const gchar *filename,
  * 
  * See the C library manual for more details about freopen().
  *
- * Returns: A <typename>FILE</typename> pointer if the file was successfully
+ * Returns: A <type>FILE</type> pointer if the file was successfully
  *    opened, or %NULL if an error occurred.
  * 
  * Since: 2.6
